#include <cassert>
#include <iostream>
#include <Eigen/Dense>
#include <nag.h>
#include <stdio.h>
#include <math.h>

struct data {
  Integer m;
  Integer n;
  double *X;
  double *y;
};

double f(const Integer &n, const double *p, const double *x){
    double res = 0.0;

    for (Integer i=0; i<n; ++i){
        res +=p[i]*x[i];
    }
    res = pow(res,2);
    return res;
}

double E(const double *p, const struct data *par){
    Integer m = par->m;
    Integer n = par->n;
    double tmp;
    double res = 0.0;
    for (Integer i = 0; i < m; i++) {
        tmp = f(n, p, &par->X[i*n]);
        res += pow(tmp-par->y[i],2);
    }
    return res;
}

void dEdp(double *p,
          const struct data *par,
          double *Jac){
    Integer n = par->n;

    double fvalm, fvalp;
    double h;


    for (Integer i = 0; i < n; i++){
        h=fabs(p[i])<1 ? sqrt(sqrt(std::numeric_limits<double>::epsilon()))
               : sqrt(sqrt(std::numeric_limits<double>::epsilon())*fabs(p[i]));
        double tmp = p[i];
        p[i] -=h;
        fvalm = E(p, par);

        p[i] = tmp + h;
        fvalp = E(p,par);
        Jac[i] = (fvalp - fvalm)/(2*h);
        p[i] = tmp;
    }

}
void fcn(Integer n, const double x[], double fvec[],
                         double fjac[], Nag_Comm *comm, Integer *iflag)
{
  struct data *par = (struct data*)comm->p;
  //Integer m = par->m;
  double h;
  double *x_tmp = new double[n];
  double *fvecm = new double[n];
  double *fvecp = new double[n];

  for (Integer i = 0; i < n; ++i) x_tmp[i] = x[i];

  if (*iflag == 1)
    dEdp(x_tmp, par, fvec);

  if (*iflag == 2){
      for (Integer i = 0; i < n; ++i){
          h=fabs(x[i])<1 ? sqrt(sqrt(std::numeric_limits<double>::epsilon()))
                         : sqrt(sqrt(std::numeric_limits<double>::epsilon())*fabs(x[i]));
          x_tmp[i] -= h;
          dEdp(x_tmp, par, fvecm);
          x_tmp[i] = x[i] + h;
          dEdp(x_tmp, par, fvecp);
          for (Integer j= 0; j < n; ++j)
              fjac[i*n + j] = (fvecp[j] - fvecm[j])/(2*h);
          x_tmp[i] = x[i];
      }
  }
  delete [] x_tmp;
  delete [] fvecm;
  delete [] fvecp;
}

int main(int argc, char* argv[])
{
  assert(argc==3); Integer m=std::stoi(argv[1]), n=std::stoi(argv[2]);

  Eigen::Matrix<double, Eigen::Dynamic, 1> y;
  Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic> t;
  /* this is the data to be fitted */
  t = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Random(m,n);
  y = Eigen::Matrix<double, Eigen::Dynamic, 1>::Random(m);
  t=t.cwiseProduct(t);

  double res;
  double *Y = new double[m];
  for (Integer i=0; i<m; i++) Y[i] = y(i);

  double *tt = new double[m*n];
  for (Integer i=0; i<m; i++){
      for (Integer j=0; j<n; j++)
          tt[i*n+j] = t(i,j);
  }

  struct data par = {m, n, tt, Y};

  /* Nag Types */
  NagError fail;
  Nag_Comm comm;
  /* For communication with user-supplied functions: */
  comm.p = &par;

  Integer exit_status = 0, i;
  double *fjac = new double[n*n];
  double *fvec = new double[n];
  double *x = new double[n];
  double xtol;

  INIT_FAIL(fail);

  printf("nag_roots_sys_deriv_easy (c05rbc) Example Program Results\n");


  /* The following starting values provide a rough solution. */
  for (i = 0; i < n; i++)
    x[i] = -1.0;

  /* nag_machine_precision (x02ajc).
   * The machine precision
   */
  xtol = sqrt(nag_machine_precision);

  /* nag_roots_sys_deriv_easy (c05rbc).
   * Solution of a system of nonlinear equations (using first
   * derivatives)
   */
  nag_roots_sys_deriv_easy(fcn, n, x, fvec, fjac, xtol, &comm, &fail);

  if (fail.code != NE_NOERROR) {
      printf("Error from nag_roots_sys_deriv_easy (c05rbc).\n%s\n",
             fail.message);
      exit_status = 1;
      if (fail.code != NE_TOO_MANY_FEVALS &&
              fail.code != NE_TOO_SMALL && fail.code != NE_NO_IMPROVEMENT)
          return exit_status;
  }

  printf(fail.code == NE_NOERROR ? "Final approximate" : "Approximate");
  printf(" solution\n\n");
  for (i = 0; i < n; i++)
      printf("%12.4f%s", x[i], (i % 3 == 2 || i == n - 1) ? "\n" : " ");

  printf("E'^T= ");
  res = 0.0;
  for (i = 0; i < n; i++){
      printf("%12.4f%s", fvec[i], (i % 3 == 2 || i == n - 1) ? "\n" : " ");
      res += fvec[i]*fvec[i];
  }
  std::cout << "||E'|| = " << sqrt(res) << std::endl;
  if (fail.code != NE_NOERROR)
      exit_status = 2;

  delete [] tt;
  delete [] Y;
  delete [] fjac;
  delete [] fvec;
  delete [] x;
  return exit_status;
}


